import enum

import Bio
import numpy as np

# Enum for different typs of coordinate system
# DO NOT changes the numbers, as the number value is stored as a feature in files
CoordinateSystem = enum.Enum(
    "CoordinateSystem", {"spherical": 1, "cubed_sphere": 2, "cartesian": 3}
)
ZDirection = enum.Enum("ZDirection", {"sidechain": 1, "backbone": 2, "outward": 3})


def define_coordinate_system(pos_N, pos_CA, pos_C, z_direction):
    """Defines a local reference system based on N, CA, and C atom positions"""

    # Define local coordinate system
    e1 = pos_C - pos_N
    e1 /= np.linalg.norm(e1)

    # Define CB positions by rotating N atoms around CA-C axis 120 degr
    pos_N_res = pos_N - pos_CA
    axis = pos_CA - pos_C
    pos_CB = np.dot(
        Bio.PDB.rotaxis((120.0 / 180.0) * np.pi, Bio.PDB.vectors.Vector(axis)),
        pos_N_res,
    )
    e2 = pos_CB
    e2 /= np.linalg.norm(e2)
    e3 = np.cross(e1, e2)

    # N-C and e2 are not perfectly perpendical to one another. We adjust e2.
    e2 = np.cross(e1, -e3)

    if z_direction == ZDirection.outward:
        # Use e3 as z-direction
        rot_matrix = np.array([e1, e2, e3])
    elif z_direction == ZDirection.backbone:
        # Use backbone direction as z-direction
        rot_matrix = np.array([e2, e3, e1])
    elif z_direction == ZDirection.sidechain:
        # Use sidechain direction as z-direction
        rot_matrix = np.array([e3, e1, e2])
    else:
        raise "Unknown z-direction "

    return rot_matrix



"""
Implements a wrapper class for the CBas VAE loading implementation.
Only minor modifications conducted by second listed authors, all credit to this herculean task go to SB

Creator: SB
Last Changes: RM
"""

__author__ = "Simon Bartels, Richard Michael"


from collections import OrderedDict
from pathlib import Path

import numpy as np
import torch
import torch.nn as nn
from scipy.stats import norm as StandardNormal
from scipy.stats.qmc import Sobol
from torch.nn import functional as F

from poli.objective_repository.gfp_cbas.abstract_vae_wrapper import AbstractVAEWrapper
from poli.objective_repository.gfp_cbas.make_vae import build_vae


class CBASVAEWrapper(AbstractVAEWrapper):
    def __init__(self, AA: int, L: int):
        self.L = L
        self.AA = AA
        model_path = Path(__file__).parent.resolve() / "assets" / "models" / "vae"
        vae_0 = build_vae(
            latent_dim=20,
            n_tokens=self.AA,  # 20,  # TODO: test if this is self.AA?
            seq_length=self.L,
            enc1_units=50,
        )
        # unfortunately the call below is not usable
        # vae_0.load_all_weights()
        print(f"Model Path: {model_path}")
        vae_suffix = "_5k_1"
        vae_0.encoder_.load_weights(
            str(model_path / f"vae_0_encoder_weights{vae_suffix}.h5")
        )
        vae_0.decoder_.load_weights(
            str(model_path / f"vae_0_decoder_weights{vae_suffix}.h5")
        )
        vae_0.vae_.load_weights(str(model_path / f"vae_0_vae_weights{vae_suffix}.h5"))

        self.vae = vae_0
        enc1_units = 50  # according to the CBAS paper
        self._decoder = ConvertedTorchVaeDecoder(
            vae_0, self.AA, self.L, vae_0.latentDim_, enc1_units
        )
        self._encoder = ConvertedTorchVaeEncoder(
            vae_0, self.AA, self.L, vae_0.latentDim_, enc1_units
        )

    def encode(self, x, grad=False):
        # seems like also for this VAE, the dimension of the amino acids is 1
        x_ = torch.flatten(
            torch.permute(torch.reshape(x, [x.shape[0], self.L, self.AA]), [0, 2, 1]),
            start_dim=1,
        )
        if not grad:
            with torch.no_grad():
                # x = self.vae.decoder_.predict(z.numpy())
                return self._encoder(x_)[0]
        else:
            return self._encoder(x_)[0]

    def decode(self, z, grad=False):
        if not grad:
            with torch.no_grad():
                # x = self.vae.decoder_.predict(z.numpy())
                return self._decoder(z)
        else:
            return self._decoder(z)

    def transform_sequences_to_atoms(self, x: torch.Tensor) -> torch.Tensor:
        """
        Turns a sequence into an approximately Dirac distribution around the sequence.
        :param x:
        :type x:
        :return:
        :rtype:
        """
        raise NotImplementedError("decide on how to handle zeros!")

    def get_max_llh_sequence(self, proba):
        assert len(proba.shape) == 3
        mle_x = F.one_hot(proba.argmax(dim=-1), self.AA)
        return mle_x.flatten(start_dim=1)

    def get_latent_dimensionality(self):
        return self.vae.latentDim_

    def decode_samples_from_encoder(
        self, x: torch.Tensor, num_samples: int
    ) -> torch.Tensor:
        s = Sobol(d=self.get_latent_dimensionality())
        samples = s.random(num_samples)
        samples = StandardNormal.ppf(samples.flatten()).reshape(samples.shape)
        with torch.no_grad():
            x_ = torch.flatten(x, start_dim=1)
            mu, logvar = self._encoder(x_)
            samples = torch.tensor(samples * np.exp(logvar.numpy() / 2) + mu.numpy())
            return (self.decode(samples, grad=False)).flatten(start_dim=1)


class ConvertedTorchVaeDecoder(torch.nn.Module):
    """
    Module that mimics the keras decoder as specified in the CBAS paper.
    """

    def __init__(self, vae, AA, L, latent_dimensionality, enc1_units):
        super().__init__()
        self.first_part = torch.nn.Sequential(
            OrderedDict(
                [
                    ("d1", torch.nn.Linear(latent_dimensionality, enc1_units)),
                    ("d1act", torch.nn.ELU()),
                    ("d3", torch.nn.Linear(enc1_units, AA * L)),
                ]
            )
        )
        named_children = dict(self.first_part.named_children())
        kt_layers = {layer.name: layer for layer in vae.decoder_.layers}
        for n, u in named_children.items():
            if n not in kt_layers.keys():
                continue
            u.weight.data = torch.as_tensor(
                kt_layers[n].get_weights()[0], dtype=torch.float64
            ).T
            u.bias.data = torch.as_tensor(
                kt_layers[n].get_weights()[1], dtype=torch.float64
            )
        self.d5 = torch.nn.Linear(AA * L, AA)
        self.d5.weight.data = torch.as_tensor(
            kt_layers["d5"].get_weights()[0], dtype=torch.float64
        ).T
        self.d5.bias.data = torch.as_tensor(
            kt_layers["d5"].get_weights()[1], dtype=torch.float64
        )

        self.AA = AA
        self.L = L

    def forward(self, z):
        return torch.nn.Softmax(dim=-1)(
            self.d5(self.first_part(z).reshape([z.shape[0], self.L, self.AA]))
        )


class ConvertedTorchVaeEncoder(torch.nn.Module):
    def __init__(self, vae, AA, L, latent_dimensionality, enc1_units):
        super().__init__()
        self.first_part = nn.Sequential(
            OrderedDict(
                [
                    # fully connected
                    ("e2", torch.nn.Linear(AA * L, enc1_units)),
                    ("e2act", torch.nn.ELU()),
                ]
            )
        )
        self.mu = torch.nn.Linear(enc1_units, latent_dimensionality)
        self.log_var = torch.nn.Linear(enc1_units, latent_dimensionality)
        named_children = dict(self.first_part.named_children())
        w = vae.vae_
        kt_layers = {layer.name: layer for layer in w.layers}
        named_children["e2"].weight.data = torch.as_tensor(
            kt_layers["e2"].get_weights()[0], dtype=torch.float64
        ).T
        named_children["e2"].bias.data = torch.as_tensor(
            kt_layers["e2"].get_weights()[1], dtype=torch.float64
        )
        self.mu.weight.data = torch.as_tensor(
            kt_layers["mu_z"].get_weights()[0], dtype=torch.float64
        ).T
        self.mu.bias.data = torch.as_tensor(
            kt_layers["mu_z"].get_weights()[1], dtype=torch.float64
        )
        self.log_var.weight.data = torch.as_tensor(
            kt_layers["log_var_z"].get_weights()[0], dtype=torch.float64
        ).T
        self.log_var.bias.data = torch.as_tensor(
            kt_layers["log_var_z"].get_weights()[1], dtype=torch.float64
        )

    def forward(self, x):
        t = self.first_part(x)
        return self.mu(t), self.log_var(t)



import glob
import os
import random
import subprocess

import numpy as np
import pandas as pd
import torch
from Bio.PDB.Polypeptide import index_to_one
from cavity_model import CavityModel, DownstreamModel, ResidueEnvironmentsDataset
from helpers import (
    compute_pdb_combo_corrs,
    ds_pred,
    ds_train_val,
    fermi_transform,
    init_lin_weights,
    populate_dfs_with_resenvs,
    remove_disulfides,
    train_loop,
    train_val_split_cavity,
    train_val_split_ds,
)
from visualization import (
    hist_plot_all,
    homology_plot,
    learning_curve_ds_with_errorbars,
    loss_analysis,
    plot_gnomad_clinvar,
    scatter_plots,
)

# Set fixed seed
seed = 0
torch.manual_seed(seed)
np.random.seed(seed)
random.seed(seed)
torch.cuda.manual_seed_all(seed)
torch.backends.cudnn.deterministic = True
torch.backends.cudnn.benchmark = False
torch.backends.cudnn.enabled = False

# Cavity constants
DEVICE = "cuda" if torch.cuda.is_available() else "cpu"
TRAIN_VAL_SPLIT_CAVITY = 0.9
BATCH_SIZE_CAVITY = 100
SHUFFLE_PDBS_CAVITY = True
LEARNING_RATE_CAVITY = 3e-4
EPOCHS_CAVITY = 20
PATIENCE_CUTOFF = 6

# Downstream constants
BATCH_SIZE_DS = 40
LEARNING_RATE_DS = 5e-4
EPOCHS_DS = 20
NUM_ENSEMBLE = 10


def main():
    # Pre-process all protein structures
    print("Pre-processing PDBs ...")
    pdb_dirs = [
        f"{os.path.dirname(os.getcwd())}/data/train/cavity/structure/",
        f"{os.path.dirname(os.getcwd())}/data/train/downstream/structure/",
        f"{os.path.dirname(os.getcwd())}/data/test/ProTherm/homology_models/structure/",
        f"{os.path.dirname(os.getcwd())}/data/test/ProTherm/targets/structure/",
        f"{os.path.dirname(os.getcwd())}/data/test/Protein_G/structure/",
        f"{os.path.dirname(os.getcwd())}/data/test/Rosetta_10/structure/",
        f"{os.path.dirname(os.getcwd())}/data/test/VAMP/structure/",
        f"{os.path.dirname(os.getcwd())}/data/test/Xstal_vs_AF2/structure/",
    ]
    for pdb_dir in pdb_dirs:
        subprocess.call(
            [
                f"{os.path.dirname(os.getcwd())}/src/pdb_parser_scripts/parse_pdbs_pred.sh",
                str(pdb_dir),
            ]
        )
    print("Pre-processing finished.")

    # Load parsed PISCES PDBs and perform train/val split
    pdb_filenames_cavity = sorted(
        glob.glob(
            f"{os.path.dirname(os.getcwd())}/data/train/cavity/structure/parsed/*coord*"
        )
    )

    if SHUFFLE_PDBS_CAVITY:
        random.shuffle(pdb_filenames_cavity)

    (
        dataloader_train_cavity,
        dataset_train_cavity,
        dataloader_val_cavity,
        dataset_val_cavity,
    ) = train_val_split_cavity(
        pdb_filenames_cavity, TRAIN_VAL_SPLIT_CAVITY, BATCH_SIZE_CAVITY, DEVICE
    )

    # Define cavity model
    cavity_model_net = CavityModel(get_latent=False).to(DEVICE)
    loss_cavity = torch.nn.CrossEntropyLoss()
    optimizer_cavity = torch.optim.Adam(
        cavity_model_net.parameters(), lr=LEARNING_RATE_CAVITY
    )

    # Train cavity model
    print("Starting cavity model training")
    best_cavity_model_path = train_loop(
        dataloader_train_cavity,
        dataloader_val_cavity,
        cavity_model_net,
        loss_cavity,
        optimizer_cavity,
        EPOCHS_CAVITY,
        PATIENCE_CUTOFF,
    )
    print("Finished cavity model training")

    # Create temporary residue environment datasets to more easily match ddG data
    pdb_filenames_ds = sorted(
        glob.glob(
            f"{os.path.dirname(os.getcwd())}/data/train/downstream/structure/parsed/*coord*"
        )
    )
    dataset = ResidueEnvironmentsDataset(pdb_filenames_ds, transformer=None)
    resenv_dataset = {}
    for resenv in dataset:
        key = (
            f"{resenv.pdb_id}{resenv.chain_id}_{resenv.pdb_residue_number}"
            f"{index_to_one(resenv.restype_index)}"
        )
        resenv_dataset[key] = resenv

    # Load ddG data to dataframe
    df_ddg = pd.read_csv(
        f"{os.path.dirname(os.getcwd())}/data/train/downstream/ddG_Rosetta/ddg.csv"
    )

    # Populate dataframes with wt ResidueEnvironment objects and wt and mt restype indices
    populate_dfs_with_resenvs(df_ddg, resenv_dataset)

    # Remove disulfides
    n_df_start = len(df_ddg)
    df_ddg = remove_disulfides(df_ddg)
    print(
        f"{n_df_start-len(df_ddg)} data points removed due to disulfides in training and validation data."
    )

    # Do Fermi transform
    df_ddg["score_fermi"] = df_ddg["score"].apply(fermi_transform)

    # Checkpoint - save
    df_ddg.to_pickle(f"{os.path.dirname(os.getcwd())}/output/ddg_train_val.pkl")
    f = open(
        f"{os.path.dirname(os.getcwd())}/output/cavity_models/best_model_path.txt", "w"
    )
    f.write(best_cavity_model_path)
    f.close()

    # Checkpoint - load
    df_ddg = pd.read_pickle(f"{os.path.dirname(os.getcwd())}/output/ddg_train_val.pkl")
    best_cavity_model_path = (
        open(
            f"{os.path.dirname(os.getcwd())}/output/cavity_models/best_model_path.txt",
            "r",
        )
        .read()
        .strip()
    )

    # Split into train and val
    dataloader_train_ds, dataloader_val_ds = train_val_split_ds(
        df_ddg, BATCH_SIZE_DS, DEVICE
    )

    # Define model
    cavity_model_net = CavityModel(get_latent=True).to(DEVICE)
    cavity_model_net.load_state_dict(
        torch.load(f"{best_cavity_model_path}", map_location=torch.device(DEVICE))
    )
    cavity_model_net.eval()
    loss_ds = torch.nn.L1Loss()

    # Train downstream model ensemble
    print("Starting downstream model ensemble training")
    for i in range(NUM_ENSEMBLE):
        # Initialize model with fixed seed
        model_idx = i
        ds_model_net = DownstreamModel().to(DEVICE)
        torch.manual_seed(seed=model_idx)
        torch.cuda.manual_seed(seed=model_idx)
        ds_model_net.apply(init_lin_weights)
        optimizer_ds = torch.optim.Adam(ds_model_net.parameters(), lr=LEARNING_RATE_DS)
        # Train model
        print(f"Training model: {model_idx+1}/{NUM_ENSEMBLE}")
        ds_train_val(
            df_ddg,
            dataloader_train_ds,
            dataloader_val_ds,
            cavity_model_net,
            ds_model_net,
            loss_ds,
            optimizer_ds,
            model_idx,
            EPOCHS_DS,
            DEVICE,
        )
        print(f"Finished training model: {model_idx+1}/{NUM_ENSEMBLE}")
    print("Finished downstream model ensemble training")

    # Plot learning curve with error bars
    learning_curve_ds_with_errorbars()

    # Load structure data
    pdb_filenames_test = {
        "Rosetta": sorted(
            glob.glob(
                f"{os.path.dirname(os.getcwd())}/data/test/Rosetta_10/structure/parsed/*coord*"
            )
        ),
        "Protein_G": sorted(
            glob.glob(
                f"{os.path.dirname(os.getcwd())}/data/test/Protein_G/structure/parsed/*coord*"
            )
        ),
        "ProTherm": sorted(
            glob.glob(
                f"{os.path.dirname(os.getcwd())}/data/test/ProTherm/targets/structure/parsed/*coord*"
            )
        ),
        "VAMP": sorted(
            glob.glob(
                f"{os.path.dirname(os.getcwd())}/data/test/VAMP/structure/parsed/*coord*"
            )
        ),
        "Homology": sorted(
            glob.glob(
                f"{os.path.dirname(os.getcwd())}/data/test/ProTherm/homology_models/structure/parsed/*coord*"
            )
        ),
        "Xstal_vs_AF2": sorted(
            glob.glob(
                f"{os.path.dirname(os.getcwd())}/data/test/Xstal_vs_AF2/structure/parsed/*coord*"
            )
        ),
    }

    # Load Rosetta ddGs
    rosetta_dict_test = {
        "Rosetta": pd.read_csv(
            f"{os.path.dirname(os.getcwd())}/data/test/Rosetta_10/ddG_Rosetta/ddg.csv"
        ),
        "Protein_G": pd.read_csv(
            f"{os.path.dirname(os.getcwd())}/data/test/Protein_G/ddG_Rosetta/ddg.csv"
        ),
        "ProTherm": pd.read_csv(
            f"{os.path.dirname(os.getcwd())}/data/test/ProTherm/targets/ddG_Rosetta/ddg.csv"
        ),
        "VAMP": pd.read_csv(
            f"{os.path.dirname(os.getcwd())}/data/test/VAMP/ddG_Rosetta/ddg.csv"
        ),
        "Homology": pd.read_csv(
            f"{os.path.dirname(os.getcwd())}/data/test/ProTherm/homology_models/ddG_Rosetta/ddg.csv"
        ),
        "Xstal_vs_AF2": pd.read_csv(
            f"{os.path.dirname(os.getcwd())}/data/test/Xstal_vs_AF2/ddG_Rosetta/ddg.csv"
        ),  # tmp
    }

    # Load experimental ddGs and VAMP-seq score data
    exp_dict_test = {
        "Protein_G": pd.read_csv(
            f"{os.path.dirname(os.getcwd())}/data/test/Protein_G/ddG_experimental/ddg.csv"
        ),
        "ProTherm": pd.read_csv(
            f"{os.path.dirname(os.getcwd())}/data/test/ProTherm/targets/ddG_experimental/ddg.csv"
        ),
        "VAMP": pd.read_csv(
            f"{os.path.dirname(os.getcwd())}/data/test/VAMP/VAMP/VAMP.csv"
        ),
        "Homology": pd.read_csv(
            f"{os.path.dirname(os.getcwd())}/data/test/ProTherm/homology_models/ddG_experimental/ddg.csv"
        ),
    }

    # Match structure, Rosetta and experimental data
    total_dict_test = {}

    for dataset_key, pdb_filenames in pdb_filenames_test.items():
        # Create temporary residue environment datasets to more easily match ddG data
        dataset_structure = ResidueEnvironmentsDataset(pdb_filenames, transformer=None)
        resenv_dataset = {}
        for resenv in dataset_structure:
            key = (
                f"{resenv.pdb_id}{resenv.chain_id}_{resenv.pdb_residue_number}"
                f"{index_to_one(resenv.restype_index)}"
            )
            resenv_dataset[key] = resenv

        # Populate Rosetta dataframes with wt ResidueEnvironment objects and wt and mt restype indices
        df_rosetta = rosetta_dict_test[dataset_key]
        df_rosetta = df_rosetta.rename(columns={"score": "score_rosetta"})
        n_rosetta_start = len(df_rosetta)
        populate_dfs_with_resenvs(df_rosetta, resenv_dataset)
        print(
            f"{n_rosetta_start-len(df_rosetta)} data points dropped when (inner) matching Rosetta and structure in: {dataset_key} data set."
        )

        # Add experimental ddGs to total dataframe
        if dataset_key in exp_dict_test.keys():
            df_exp = exp_dict_test[dataset_key]
            df_exp = df_exp.rename(columns={"score": "score_exp"})
            if dataset_key == "Homology":
                df_rosetta["pdbid_target"] = df_rosetta["pdbid"].str[:4]
                df_exp["pdbid_target"] = df_exp["pdbid"].str[:4]
                df_rosetta = df_rosetta.merge(
                    df_exp, on=["pdbid_target", "chainid", "variant"], how="inner"
                )
                df_rosetta["pdbid"] = df_rosetta["pdbid_x"]
                df_rosetta.drop(["pdbid_x", "pdbid_y"], axis=1)
            else:
                df_rosetta = df_rosetta.merge(
                    df_exp, on=["pdbid", "chainid", "variant"], how="inner"
                )
            print(
                f"{len(df_exp)-len(df_rosetta)} data points dropped when (inner) matching Rosetta and structural data with experimental data in: {dataset_key} data set."
            )
        df_total = df_rosetta

        # Remove disulfides
        if dataset_key == "Rosetta":
            n_df_start = len(df_total)
            df_total = remove_disulfides(df_total)
            print(
                f"{n_df_start-len(df_total)} data points removed due to disulfides in: {dataset_key} data set."
            )

        # Do Fermi transform
        df_total["score_rosetta_fermi"] = df_total["score_rosetta"].apply(
            fermi_transform
        )
        if "score_exp" in df_total:
            df_total["score_exp_fermi"] = df_total["score_exp"].apply(fermi_transform)

        # Add to total dict
        total_dict_test[dataset_key] = df_total

    # Add Protein G to ProTherm data set
    total_dict_test["ProTherm_with_Protein_G"] = pd.concat(
        [total_dict_test["ProTherm"], total_dict_test["Protein_G"]], axis=0
    )

    # Initialize models
    cavity_model_net = CavityModel(get_latent=True).to(DEVICE)
    best_cavity_model_path = (
        open(
            f"{os.path.dirname(os.getcwd())}/output/cavity_models/best_model_path.txt",
            "r",
        )
        .read()
        .strip()
    )
    cavity_model_net.load_state_dict(
        torch.load(f"{best_cavity_model_path}", map_location=torch.device(DEVICE))
    )
    cavity_model_net.eval()
    ds_model_net = DownstreamModel().to(DEVICE)

    # Make test set predictions
    for dataset_key, df_total in total_dict_test.items():
        print(f"Computing predictions for data set: {dataset_key}")

        # Compute predictions
        df_ml = ds_pred(
            cavity_model_net, ds_model_net, df_total, dataset_key, NUM_ENSEMBLE, DEVICE
        )

        # Merge and save data with predictions
        df_total = df_total.merge(
            df_ml, on=["pdbid", "chainid", "variant"], how="inner"
        )
        if dataset_key != "Homology":
            print(
                f"{len(df_ml)-len(df_total)} data points dropped when (inner) matching total data with ml predictions in: {dataset_key} data set."
            )
        df_total.to_csv(
            f"{os.path.dirname(os.getcwd())}/output/{dataset_key}/ml_preds.csv",
            index=False,
        )

        # Make visualizations
        if dataset_key == "Xstal_vs_AF2":
            compute_pdb_combo_corrs(df_total, dataset_key)
        else:
            scatter_plots(df_total, dataset_key)
            loss_analysis(df_total, dataset_key)
            if dataset_key == "Homology":
                homology_plot(df_total, dataset_key)
        print(f"Finished making predictions for data set: {dataset_key}")

    # Plot large cytosolic data set
    filenames = sorted(
        glob.glob(f"{os.path.dirname(os.getcwd())}/data/test/Cytosolic/*.csv")
    )
    df = pd.DataFrame()
    for file in filenames:
        df_file = pd.read_csv(file)
        df = pd.concat((df, df_file), axis=0)
    hist_plot_all(df, "Cytosolic")

    # Plot large cytosolic data set vs gnomAD and ClinVar
    plot_gnomad_clinvar()


if __name__ == "__main__":
    main()



"""This module contains the default values for the amino acid encoding.

For amino acid encoding, we use the following:

```python
AMINO_ACIDS = [
    "A",
    "R",
    "N",
    "D",
    "C",
    "E",
    "Q",
    "G",
    "H",
    "I",
    "L",
    "K",
    "M",
    "F",
    "P",
    "S",
    "T",
    "W",
    "Y",
    "V",
]
```

"""

AMINO_ACIDS = [
    "A",
    "R",
    "N",
    "D",
    "C",
    "E",
    "Q",
    "G",
    "H",
    "I",
    "L",
    "K",
    "M",
    "F",
    "P",
    "S",
    "T",
    "W",
    "Y",
    "V",
]

ENCODING = {amino_acid: i for i, amino_acid in enumerate(AMINO_ACIDS)}



from __future__ import annotations

import numpy as np


def _construct_banded_matrix(size: int, band_length: int | None = None) -> np.ndarray:
    """
    Constructs a matrix of zeroes and ones, where
    the ones are bands that can loop around.

    The length of the non-zero band is given by band_length.
    By default, it is set to size - 2 * (size // 5). This
    value is taken from the original Ehrlich paper [1].

    References
    ----------
    [1] Stanton, S., Alberstein, R., Frey, N., Watkins, A., & Cho, K. (2024).
    Closed-Form Test Functions for Biophysical Sequence Optimization Algorithms.
    arXiv preprint arXiv:2407.00236. https://arxiv.org/abs/2407.00236
    """
    matrix = np.zeros((size, size), dtype=int)
    band_index = 0
    band_length = size - ((2 * size) // 5) if band_length is None else band_length
    for row_i in range(size):
        indices_for_positions_that_will_be_1 = list(
            range(band_index, band_index + band_length)
        )

        # Looping the ones that go over the limit
        for i in range(len(indices_for_positions_that_will_be_1)):
            if indices_for_positions_that_will_be_1[i] >= size:
                indices_for_positions_that_will_be_1[i] -= size

        matrix[row_i, indices_for_positions_that_will_be_1] = 1

        band_index += 1

    return matrix


def _construct_binary_mask(size: int, band_length: int | None = None) -> np.ndarray:
    banded_matrix = _construct_banded_matrix(size, band_length=band_length)

    # Shuffle its rows
    random_indices_for_rows = np.random.permutation(size)
    binary_mask_matrix = banded_matrix[random_indices_for_rows]

    # Making sure that the diagonal is full
    # of ones
    binary_mask_matrix[np.diag_indices(size)] = 1

    return binary_mask_matrix


def _construct_transition_matrix(
    size: int,
    seed: int | None = None,
    temperature: float = 0.5,
    band_length: int | None = None,
) -> np.ndarray:
    binary_mask_matrix = _construct_binary_mask(size, band_length=band_length)

    # Creating a random state and matrix
    random_state = np.random.RandomState(seed)
    random_matrix = random_state.randn(size, size)

    # Softmax it with low temperature
    transition_matrix = np.exp(random_matrix / temperature) / np.sum(
        np.exp(random_matrix / temperature), axis=0
    )

    # Mask it
    masked_transition_matrix = transition_matrix * binary_mask_matrix

    # Normalize it
    normalized_transition_matrix = masked_transition_matrix / np.sum(
        masked_transition_matrix, axis=1, keepdims=True
    )

    return normalized_transition_matrix



"""A wrapper around observers that prohibits the use of observer.observe"""

from poli.core.util.abstract_observer import AbstractObserver


class AlgorithmObserverWrapper:
    def __init__(self, observer: AbstractObserver):
        self._observer = observer

    def log(self, algorithm: dict):
        self._observer.log(algorithm)



"""
Test suite for benchmarks.

This test suite only verifies that the benchmarks can
be created without errors. It does not check the correctness
of the benchmarks. In the case of PMO and Guacamol, those
tests are already covered by the tests in objective_repository.
"""

import pytest


def test_creating_toy_continuous_functions_benchmark():
    from poli.benchmarks import ToyContinuousFunctionsBenchmark

    benchmark = ToyContinuousFunctionsBenchmark()

    for problem in benchmark:
        f, x0 = problem.black_box, problem.x0
        f(x0)


def test_creating_embedded_toy_continuous_functions_benchmark():
    from poli.benchmarks import EmbeddedBranin2D, EmbeddedHartmann6D

    for benchmark in [EmbeddedBranin2D(), EmbeddedHartmann6D()]:
        for problem in benchmark:
            f, x0 = problem.black_box, problem.x0
            f(x0)


def test_names_from_guacamol_benchmark():
    from poli.benchmarks import GuacaMolGoalDirectedBenchmark

    assert GuacaMolGoalDirectedBenchmark(
        string_representation="SMILES"
    ).problem_names == [
        "albuterol_similarity",
        "amlodipine_mpo",
        "celecoxib_rediscovery",
        "deco_hop",
        "fexofenadine_mpo",
        "isomer_c7h8n2o2",
        "isomer_c9h10n2o2pf2cl",
        "median_1",
        "median_2",
        "mestranol_similarity",
        "osimetrinib_mpo",
        "perindopril_mpo",
        "ranolazine_mpo",
        "rdkit_logp",
        "rdkit_qed",
        "sa_tdc",
        "scaffold_hop",
        "sitagliptin_mpo",
        "thiothixene_rediscovery",
        "troglitazone_rediscovery",
        "valsartan_smarts",
        "zaleplon_mpo",
    ]


@pytest.mark.poli__tdc
def test_creating_guacamol_benchmark():
    from poli.benchmarks import GuacaMolGoalDirectedBenchmark

    benchmark = GuacaMolGoalDirectedBenchmark(string_representation="SELFIES")

    for problem in benchmark:
        _, _ = problem.black_box, problem.x0

        # Break after the first iteration
        # for CI efficiency
        break


@pytest.mark.poli__tdc
def test_creating_pmo_benchmark():
    from poli.benchmarks import PMOBenchmark

    benchmark = PMOBenchmark(string_representation="SELFIES")

    for problem in benchmark:
        _, _ = problem.black_box, problem.x0

        # Break after the first iteration
        # for CI efficiency. The creation of all
        # these black boxes is already being tested
        # in the chemistry test suite.
        break



"""
Implements an abstract TDC black box for all the PMO [1, 2] related problems.

PMO stands for Practical Molecular Optimization, a benchmark
suite for molecular optimization problems that extends GuacaMol [3].

References
----------
[1] Artificial intelligence foundation for therapeutic science.
    Huang, K., Fu, T., Gao, W. et al.  Nat Chem Biol 18, 1033-1036 (2022). https://doi.org/10.1038/s41589-022-01131-2
[2] Sample Efficiency Matters: A Benchmark for Practical Molecular Optimization
    Wenhao Gao, Tianfan Fu, Jimeng Sun, Connor W. Coley
    https://arxiv.org/abs/2206.12411
[3] GuacaMol: benchmarking models for de novo molecular design.
    Brown, N. et al.  J Chem Inf Model 59 (2019).
    https://pubs.acs.org/doi/10.1021/acs.jcim.8b00839
"""

from __future__ import annotations

from typing import Literal

import numpy as np

from poli.core.abstract_black_box import AbstractBlackBox
from poli.core.util.isolation.instancing import get_inner_function


class TDCBlackBox(AbstractBlackBox):
    """
    An abstract black box for the TDC (Therapeutics Data
    Commons) problems [1].

    Parameters
    ----------
    oracle_name: str
        The name of the oracle to be used.
    string_representation : Literal["SMILES", "SELFIES"], optional
        A string (either "SMILES" or "SELFIES") specifying which
        molecule representation you plan to use.
    alphabet : list[str] | None, optional
        The alphabet to be used for the SMILES or SELFIES representation.
        It is common that the alphabet depends on the dataset used, so
        it is recommended to pass it as an argument. Default is None.
    max_sequence_length : int, optional
        The maximum length of the sequence. Default is infinity.
    batch_size : int, optional
        The batch size for simultaneous execution, by default None.
    parallelize : bool, optional
        Flag indicating whether to parallelize execution, by default False.
    num_workers : int, optional
        The number of workers for parallel execution, by default None.
    evaluation_budget:  int, optional
        The maximum number of function evaluations. Default is infinity.
    force_isolation: bool, optional
        Whether to force the isolation of the black box. Default is False.
    **kwargs_for_oracle: dict
        Other keyword arguments to be passed to the oracle.

    Attributes
    ----------
    oracle_name : str
        The name of the oracle.

    Methods
    -------
    __init__(self, oracle_name, string_representation="SMILES", force_isolation=False, batch_size=None, parallelize=False, num_workers=None, evaluation_budget=float("inf"), **kwargs_for_oracle
        Initializes a new instance of the abstract TDC class.

    References
    ----------
    [1] Artificial intelligence foundation for therapeutic science.
        Huang, K., Fu, T., Gao, W. et al.  Nat Chem Biol 18, 1033-1036 (2022). https://doi.org/10.1038/s41589-022-01131-2
    """

    def __init__(
        self,
        oracle_name: str,
        string_representation: Literal["SMILES", "SELFIES"] = "SMILES",
        alphabet: list[str] | None = None,
        max_sequence_length: int = np.inf,
        force_isolation: bool = False,
        batch_size: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
        **kwargs_for_oracle,
    ):
        if parallelize:
            print(
                "poli 🧪: TDCBlackBox parallelization is handled by oracles. Disabling it."
            )
            parallelize = False
        super().__init__(
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
        )
        self.oracle_name = oracle_name
        self.alphabet = alphabet
        self.max_sequence_length = max_sequence_length
        self.string_representation = string_representation

        from_smiles = string_representation.upper() == "SMILES"
        self.inner_function = get_inner_function(
            isolated_function_name="tdc__isolated",
            class_name="TDCIsolatedFunction",
            module_to_import="poli.core.chemistry.tdc_isolated_function",
            force_isolation=force_isolation,
            oracle_name=oracle_name,
            from_smiles=from_smiles,
            **kwargs_for_oracle,
        )

    def _black_box(self, x: np.ndarray, context=None) -> np.ndarray:
        return self.inner_function(x, context)



"""Executable script used for isolation of objective factories and functions.

The equivalent of objective, but for isolated black boxes instead of problem factories.
"""

import argparse
import logging
import os
import sys
import traceback

from poli.core.abstract_isolated_function import AbstractIsolatedFunction
from poli.core.util.inter_process_communication.process_wrapper import get_connection
from poli.core.util.seeding import seed_python_numpy_and_torch

ADDITIONAL_IMPORT_SEARCH_PATHES_KEY = "ADDITIONAL_IMPORT_PATHS"


def dynamically_instantiate(obj: str, **kwargs):
    """Dynamically instantiates an object from a string.

    This function is used internally to instantiate objective
    factories dynamically, inside isolated processes. It is
    also used to instantiate external observers.

    Parameters
    ----------
    obj : str
        The string containing the name of the object to be instantiated.
    **kwargs : dict
        The keyword arguments to be passed to the object constructor.
    """
    # FIXME: this method opens up a serious security vulnerability
    # TODO: possible alternative: importlib
    # TODO: another possible alternative: hydra
    # sys.path.append(os.getcwd())
    sys.path.extend(os.environ[ADDITIONAL_IMPORT_SEARCH_PATHES_KEY].split(":"))
    # sys.path.extend(os.environ['PYTHONPATH'].split(':'))
    last_dot = obj.rfind(".")

    command = (
        "from "
        + obj[:last_dot]
        + " import "
        + obj[last_dot + 1 :]
        + " as DynamicObject"
    )
    try:
        exec(command)
        instantiated_object = eval("DynamicObject")(**kwargs)
    except ImportError as e:
        logging.fatal(f"Path: {os.environ['PATH']}")
        logging.fatal(f"Python path: {sys.path}")
        logging.fatal(f"Path: {os.environ[ADDITIONAL_IMPORT_SEARCH_PATHES_KEY]}")
        if "PYTHONPATH" in os.environ.keys():
            logging.fatal(f"Path: {os.environ['PYTHONPATH']}")
        else:
            logging.fatal("PYTHONPATH is not part of the environment variables.")
        raise e
    return instantiated_object


def run(objective_name: str, port: int, password: str) -> None:
    """Starts an objective function listener loop to wait for requests.

    Parameters
    ----------
    factory_kwargs : str
        The string containing the factory kwargs (see ProcessWrapper
        for details about how this factory_kwargs strings is built).
    objective_name : str
        The name of the objective function to be instantiated.
    port : int
        The port number for the connection with the mother process.
    password : str
        The password for the connection with the mother process.
    """
    # kwargs = parse_factory_kwargs(factory_kwargs)

    # make connection with the mother process
    conn = get_connection(port, password)

    # TODO: We could be receiving the kwargs for the factory here.
    msg_type, seed, kwargs_for_function = conn.recv()

    if seed is not None:
        seed_python_numpy_and_torch(seed)

    # dynamically load objective function module
    # At this point, the black box objective function
    # is exactly the same as the one used in the
    # registration (?).
    try:
        f: AbstractIsolatedFunction = dynamically_instantiate(
            objective_name, **kwargs_for_function
        )

        # give mother process the signal that we're ready
        conn.send(["SETUP", None])
    except Exception as e:
        tb = traceback.format_exc()
        conn.send(["EXCEPTION", e, tb])
        raise e

    # now wait for objective function calls
    while True:
        msg_type, *msg = conn.recv()

        if msg_type == "QUIT":
            break
        try:
            if msg_type == "QUERY":
                x, context = msg
                y = f(x, context=context)
                conn.send(["QUERY", y])
            elif msg_type == "ATTRIBUTE":
                attribute_name = msg[0]
                attribute = getattr(f, attribute_name)
                conn.send(["ATTRIBUTE", attribute])
            elif msg_type == "IS_METHOD":
                attribute_name = msg[0]
                is_method = callable(getattr(f, attribute_name))
                conn.send(["IS_METHOD", is_method])
            elif msg_type == "METHOD":
                method_name = msg[0]
                method_args = msg[1]
                method_kwargs = msg[2]
                method = getattr(f, method_name)
                result = method(*method_args, **method_kwargs)
                conn.send(["METHOD", result])
        except Exception as e:
            tb = traceback.format_exc()
            conn.send(["EXCEPTION", e, tb])

    # conn.close()
    # exit()  # kill other threads, and close file handles


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--objective-name", required=True)
    parser.add_argument("--port", required=True, type=int)
    parser.add_argument("--password", required=True, type=str)

    args, factory_kwargs = parser.parse_known_args()
    run(args.objective_name, args.port, args.password)



"""Abstract implementation for isolated logic.

This module contains the implementation of an abstract isolated function.
Isolated functions allow you to implement complex, heavily-dependent
objective functions inside isolated processes, which can be run in
their own conda environments. This is useful for running objective
functions that have dependencies that are not compatible with the
dependencies of the main process (e.g. your optimizers, or observers).

AbstractIsolatedFunctions are only expected to implement the __call__
method. This method will (usually) be used to evaluate the function
inside the `_black_box(x, context)` method of black box.

Diving deeper, these isolated functions are used in the `ExternalFunction`
class, which maintains a communication with an isolated process running
the `external_isolated_function_script.py` script.
"""

import numpy as np


class AbstractIsolatedFunction:
    def __init__(self) -> None:
        pass

    def __call__(self, x: np.ndarray, context=None) -> np.ndarray:
        raise NotImplementedError

    def terminate(self):
        pass



"""Utilities for interacting with the original RaSP codebase."""

from .load_models import load_cavity_and_downstream_models
from .rasp_interface import RaspInterface

__all__ = ["load_cavity_and_downstream_models", "RaspInterface"]



"""Abstract class for observers in poli."""

import numpy as np

from poli.core.black_box_information import BlackBoxInformation


class AbstractObserver:
    """
    Abstract base class for observers in the poli library.

    Observers are used to monitor the progress of optimization algorithms
    by observing the values of the objective function and the decision variables
    at each iteration.

    IMPORTANT: The __init__ function of observer classes MUST NOT require arguments.

    Methods
    -------
        observe(x: np.ndarray, y: np.ndarray, context=None) -> None:
            Observes the values of the objective function and the decision variables
            at each iteration of the optimization algorithm. If the observer is
            set in the creation of an objective function, this method will be
            called every time the objective function is called.

        initialize_observer(problem_setup_info: ProblemSetupInformation, caller_info: object, seed: int) -> object:
            Initializes the observer with the necessary information to monitor
            the optimization process.

        finish() -> None:
            Performs any necessary cleanup or finalization steps after the
            optimization process is complete.
    """

    def observe(self, x: np.ndarray, y: np.ndarray, context=None) -> None:
        """
        Observe the given data points.

        Parameters
        ----------
        x: np.ndarray
            The input data points.
        y: np.ndarray
            The output data points.
        context: object
            Additional context for the observation.

        Raises
        -------
        NotImplementedError:
            This method is meant to be overridden by subclasses.
        """
        raise NotImplementedError("abstract method")

    def initialize_observer(
        self,
        problem_setup_info: BlackBoxInformation,
        caller_info: object,
        seed: int,
    ) -> object:
        """
        Initialize the observer.

        Parameters
        ----------
        problem_setup_info : BlackBoxInformation
            The information about the problem setup.
        caller_info : object
            The information that is passed to the observer at init time.
        seed : int
            The seed value for random number generation.

        Returns
        -------
        observer: AbstractObserver (self)
            The observer object. (This way, one can chain the `initialize_observer` method.)

        Raises
        ------
        NotImplementedError
            If the method is not implemented in the derived class.

        """
        raise NotImplementedError("abstract method")

    def log(self, algorithm_info: dict):
        pass

    def finish(self) -> None:
        """Finish the observer."""
        pass

    def __del__(self):
        try:
            self.finish()
        except Exception:
            pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.finish()

    def __str__(self) -> str:
        return f"{self.__class__.__name__}"

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}>"



"""
This module implements Rapid Stability Predictions (RaSP),
using their open-source code [1].

From a bird's eye view, RaSP is a black box that starts
with a collection of wildtype pdb files, and assesses the
stability of a (single-site) mutant. For the informed reader,
this black box can be considered a drop-in replacement of FoldX,
or Rosetta.

[1] “Rapid Protein Stability Prediction Using Deep Learning Representations.”
Blaabjerg, Lasse M, Maher M Kassem, Lydia L Good, Nicolas Jonsson,
Matteo Cagiada, Kristoffer E Johansson, Wouter Boomsma, Amelie Stein,
and Kresten Lindorff-Larsen.  Edited by José D Faraldo-Gómez,
Detlef Weigel, Nir Ben-Tal, and Julian Echave. eLife 12
(May 2023): e82593. https://doi.org/10.7554/eLife.82593.

"""

from __future__ import annotations

from pathlib import Path
from typing import List, Union

from poli.core.abstract_black_box import AbstractBlackBox
from poli.core.abstract_problem_factory import AbstractProblemFactory
from poli.core.black_box_information import BlackBoxInformation
from poli.core.problem import Problem
from poli.core.util.isolation.instancing import get_inner_function
from poli.core.util.proteins.defaults import AMINO_ACIDS
from poli.core.util.seeding import seed_python_numpy_and_torch


class RaspBlackBox(AbstractBlackBox):
    """
    RaSP Black Box implementation.

    Parameters
    ----------
    wildtype_pdb_path : Union[Path, List[Path]]
        The path(s) to the wildtype PDB file(s), by default None.
    additive : bool, optional
        Whether we treat multiple mutations as additive, by default False.
        If you are interested in running this black box with multiple
        mutations, you should set this to True. Otherwise, it will
        raise an error if you pass a sequence with more than one mutation.
    chains_to_keep : List[str], optional
        The chains to keep in the PDB file(s), by default we
        keep the chain "A" for all pdbs passed.
    experiment_id : str, optional
        The experiment ID, by default None.
    tmp_folder : Path, optional
        The temporary folder path, by default None, which means
        we will keep temporary files in /tmp.
    batch_size : int, optional
        The batch size for parallel evaluation, by default None.
    parallelize : bool, optional
        Flag indicating whether to parallelize evaluation, by default False.
    num_workers : int, optional
        The number of workers for parallel evaluation, by default None.
    evaluation_budget : int, optional
        The evaluation budget, by default float("inf").

    Methods
    -------
    _black_box(x, context=None)
        The main black box method that performs the computation, i.e.
        it computes the stability of the mutant(s) in x.
    _clean_wildtype_pdb_files()
        This function cleans the wildtype pdb files
        stored in self.wildtype_pdb_paths, using
        cached results if they exist.


    Raises
    ------
    AssertionError
        If wildtype_pdb_path is not provided.

    Notes
    -----
    - The wildtype_pdb_path can be a single Path object or a list of Path objects.
    - If chains_to_keep is not provided, it defaults to keeping chain A for all wildtypes.
    - If experiment_id is not provided, it is generated using the current timestamp and a random UUID.
    - If tmp_folder is not provided, it defaults to the default temporary path.
    """

    def __init__(
        self,
        wildtype_pdb_path: Union[Path, List[Path]],
        additive: bool = False,
        chains_to_keep: List[str] = None,
        penalize_unfeasible_with: float | None = None,
        device: str | None = None,
        experiment_id: str = None,
        tmp_folder: Path = None,
        batch_size: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
        force_isolation: bool = False,
    ):
        """
        Initialize the RaSP Register object.

        Parameters:
        -----------
        wildtype_pdb_path : Union[Path, List[Path]]
            The path(s) to the wildtype PDB file(s).
        additive : bool, optional
            Whether we treat multiple mutations as additive, by default False.
            If you are interested in running this black box with multiple
            mutations, you should set this to True. Otherwise, it will
            raise an error if you pass a sequence with more than one mutation.
        chains_to_keep : List[str], optional
            The chains to keep in the PDB file(s), by default we
            keep the chain "A" for all pdbs passed.
        penalize_unfeasible_with : float | None, optional
            The value to penalize unfeasible solutions with, by default None, which means we raise an error when
            an unfeasible sequence (e.g. a sequence with a length
            different from the wildtypes) is passed.
        device: str, optional
            The device to load the models on, by default None, which
            means that RaSP decides to run on either CUDA or
            CPU, depending on the availability of CUDA.
        experiment_id : str, optional
            The experiment ID, by default None.
        tmp_folder : Path, optional
            The temporary folder path, by default None.
        batch_size : int, optional
            The batch size for parallel evaluation, by default None.
        parallelize : bool, optional
            Flag indicating whether to parallelize evaluation, by default False.
        num_workers : int, optional
            The number of workers for parallel evaluation, by default None.
        evaluation_budget : int, optional
            The evaluation budget, by default float("inf").

        Raises:
        -------
        AssertionError
            If wildtype_pdb_path is not provided.

        Notes:
        ------
        - The wildtype_pdb_path can be a single Path object or a list of Path objects.
        - If chains_to_keep is not provided, it defaults to keeping chain A for all wildtypes.
        - If experiment_id is not provided, it is generated using the current timestamp and a random UUID.
        - If tmp_folder is not provided, it defaults to the default temporary path.
        """
        assert wildtype_pdb_path is not None, (
            "Missing required argument wildtype_pdb_file. "
            "Did you forget to pass it to create and into the black box?"
        )
        if parallelize:
            print(
                "poli 🧪: RaspBlackBox parallelization is handled by the isolated logic. Disabling it."
            )
            parallelize = False
        super().__init__(
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
        )
        self.force_isolation = force_isolation
        self.wildtype_pdb_path = wildtype_pdb_path
        self.chains_to_keep = chains_to_keep
        self.experiment_id = experiment_id
        self.tmp_folder = tmp_folder
        self.additive = additive
        self.penalize_unfeasible_with = penalize_unfeasible_with
        self.device = device
        self.inner_function = get_inner_function(
            isolated_function_name="rasp__isolated",
            class_name="RaspIsolatedLogic",
            module_to_import="poli.objective_repository.rasp.isolated_function",
            force_isolation=self.force_isolation,
            wildtype_pdb_path=self.wildtype_pdb_path,
            additive=self.additive,
            chains_to_keep=self.chains_to_keep,
            penalize_unfeasible_with=self.penalize_unfeasible_with,
            experiment_id=self.experiment_id,
            tmp_folder=self.tmp_folder,
            device=self.device,
        )
        self.x0 = self.inner_function.x0

    def _black_box(self, x, context=None):
        """
        Computes the stability of the mutant(s) in x.

        Parameters
        ----------
        x : np.ndarray
            Input array of shape [b, L] containing strings.
        context : dict, optional
            Additional context information (default is None).

        Returns
        -------
        y : np.ndarray
            The stability of the mutant(s) in x.

        Notes
        -----
        - x is a np.array[str] of shape [b, L], where L is the length
          of the longest sequence in the batch, and b is the batch size.
          We process it by concantenating the array into a single string,
          where we assume the padding to be an empty string (if there was any).
          Each of these x_i's will be matched to the wildtype in self.  wildtype_residue_strings with the lowest Hamming distance.
        """
        return self.inner_function(x, context=context)

    def get_black_box_info(self) -> BlackBoxInformation:
        """
        Returns the black box information for RaSP.
        """
        is_aligned = False if len(self.wildtype_pdb_path) > 1 else True
        is_fixed_length = False if len(self.wildtype_pdb_path) > 1 else True
        max_sequence_length = max([len("".join(x)) for x in self.x0])
        return BlackBoxInformation(
            name="rasp",
            max_sequence_length=max_sequence_length,
            aligned=is_aligned,
            fixed_length=is_fixed_length,
            deterministic=True,
            alphabet=AMINO_ACIDS,
            log_transform_recommended=False,
            discrete=True,
            fidelity="low",
            padding_token="",
        )


class RaspProblemFactory(AbstractProblemFactory):
    def create(
        self,
        wildtype_pdb_path: Union[Path, List[Path]],
        additive: bool = False,
        chains_to_keep: List[str] = None,
        penalize_unfeasible_with: float | None = None,
        device: str | None = None,
        experiment_id: str = None,
        tmp_folder: Path = None,
        seed: int = None,
        batch_size: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
        force_isolation: bool = False,
    ) -> Problem:
        """
        Creates a RaSP black box instance, alongside initial
        observations.

        Parameters
        ----------
        wildtype_pdb_path : Union[Path, List[Path]]
            The path(s) to the wildtype PDB file(s).
        additive: bool, optional
            Whether we treat multiple mutations as additive, by default False.
            If you are interested in running this black box with multiple
            mutations, you should set this to True. Otherwise, it will
            raise an error if you pass a sequence with more than one mutation.
        chains_to_keep : List[str], optional
            The chains to keep in the PDB file(s), by default we
            keep the chain "A" for all pdbs passed.
        penalize_unfeasible_with : float | None, optional
            The value to penalize unfeasible solutions with, by default None,
            which means we raise an error when an unfeasible sequence (e.g.
            a sequence with a length different from the wildtypes) is passed.
        device: str, optional
            The device to load the models on, by default None,
            which means that RaSP decides to run on either CUDA or
            CPU, depending on the availability of CUDA.
        experiment_id : str, optional
            The experiment ID, by default None.
        tmp_folder : Path, optional
            The temporary folder path, by default None.
        seed : int, optional
            The seed value for random number generation, by default None.
        batch_size : int, optional
            The batch size for parallel evaluation, by default None.
        parallelize : bool, optional
            Flag indicating whether to parallelize evaluation, by default False.
        num_workers : int, optional
            The number of workers for parallel evaluation, by default None.
        evaluation_budget : int, optional
            The evaluation budget, by default float("inf").

        Returns
        -------
        f : RaspBlackBox
            The RaSP black box instance.
        x0 : np.ndarray
            The initial observations (i.e. the wildtypes as sequences
            of amino acids).
        y0 : np.ndarray
            The initial observations (i.e. the stability of the wildtypes).
        """
        if seed is not None:
            seed_python_numpy_and_torch(seed)

        if wildtype_pdb_path is None:
            raise ValueError(
                "Missing required argument wildtype_pdb_path. "
                "Did you forget to pass it to create()?"
            )

        if isinstance(wildtype_pdb_path, str):
            wildtype_pdb_path = [Path(wildtype_pdb_path.strip())]
        elif isinstance(wildtype_pdb_path, Path):
            wildtype_pdb_path = [wildtype_pdb_path]
        elif isinstance(wildtype_pdb_path, list):
            if isinstance(wildtype_pdb_path[0], str):
                wildtype_pdb_path = [Path(x.strip()) for x in wildtype_pdb_path]
            elif isinstance(wildtype_pdb_path[0], Path):
                pass
        else:
            raise ValueError(
                "Invalid type for wildtype_pdb_path. "
                "It must be a string, a Path, or a list of strings or Paths."
            )

        f = RaspBlackBox(
            wildtype_pdb_path=wildtype_pdb_path,
            additive=additive,
            chains_to_keep=chains_to_keep,
            penalize_unfeasible_with=penalize_unfeasible_with,
            device=device,
            experiment_id=experiment_id,
            tmp_folder=tmp_folder,
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
            force_isolation=force_isolation,
        )

        # Constructing x0
        # (Moved to the isolated logic)
        x0 = f.inner_function.x0

        problem = Problem(f, x0)

        return problem
